package eu.jonahbauer.json;

import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;

/**
 * Java representation of a JSON array.
 * @param elements the array's elements
 */
@SuppressWarnings("unused")
public record JsonArray(
        @NotNull List<@Nullable JsonValue> elements
) implements JsonValue, List<@Nullable JsonValue> {

    public JsonArray {
        Objects.requireNonNull(elements, "elements");
        if (!Util.isImmutable(elements)) {
            elements = elements.stream().toList();
        }
    }

    /**
     * Creates a {@link JsonArray} from the given list, using {@link JsonValue#valueOf(Object)} to convert the objects to
     * {@link JsonValue}s.
     * @param list a list of objects
     * @return a {@link JsonArray}
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(@Nullable List<?> list) {
        if (list == null) return null;
        if (list instanceof JsonArray json) return json;
        return new JsonArray(list.stream().map(JsonValue::valueOf).toList());
    }

    /**
     * Creates a {@link JsonArray} from the given array, using {@link JsonValue#valueOf(Object)} to convert the objects to
     * {@link JsonValue}s.
     * @param array an array of objects
     * @return a {@link JsonArray}
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(@Nullable Object @Nullable... array) {
        return array == null ? null : new JsonArray(Arrays.stream(array).map(JsonValue::valueOf).toList());
    }

    /**
     * Creates a {@link JsonArray} from the given array, using {@link JsonNumber#valueOf(int)} to convert the
     * {@code int}s to {@link JsonNumber}s.
     * @param array an array of {@code int}s
     * @return a {@link JsonArray}
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(int @Nullable... array) {
        return array == null ? null : new JsonArray(Arrays.stream(array).<JsonValue>mapToObj(JsonNumber::valueOf).toList());
    }

    /**
     * Creates a {@link JsonArray} from the given array, using {@link JsonNumber#valueOf(double)} to convert the
     * {@code double}s to {@link JsonNumber}s.
     * @param array an array of {@code double}s
     * @return a {@link JsonArray}
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(double @Nullable... array) {
        return array == null ? null : new JsonArray(Arrays.stream(array).<JsonValue>mapToObj(JsonNumber::valueOf).toList());
    }

    //<editor-fold desc="List" defaultstate="collapsed">
    @Override
    public int size() {
        return elements.size();
    }

    @Override
    public boolean isEmpty() {
        return elements.isEmpty();
    }

    @Override
    public boolean contains(Object o) {
        return elements.contains(o);
    }

    @Override
    public @NotNull Iterator<@Nullable JsonValue> iterator() {
        return elements.iterator();
    }

    @Override
    public Object @NotNull[] toArray() {
        return elements.toArray();
    }

    @Override
    public <T> T @NotNull[] toArray(T @NotNull[] a) {
        return elements.toArray(a);
    }

    @Override
    public boolean add(@Nullable JsonValue value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean remove(Object o) {
        throw new UnsupportedOperationException();
    }

    @Override
    @SuppressWarnings("SlowListContainsAll")
    public boolean containsAll(@NotNull Collection<?> c) {
        return elements.containsAll(c);
    }

    @Override
    public boolean addAll(@NotNull Collection<? extends @Nullable JsonValue> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean addAll(int index, @NotNull Collection<? extends @Nullable JsonValue> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean removeAll(@NotNull Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean retainAll(@NotNull Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }

    @Override
    public @Nullable JsonValue get(int index) {
        return elements.get(index);
    }

    @Override
    public @Nullable JsonValue set(int index, @Nullable JsonValue element) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void add(int index, @Nullable JsonValue element) {
        throw new UnsupportedOperationException();
    }

    @Override
    public @Nullable JsonValue remove(int index) {
        throw new UnsupportedOperationException();
    }

    @Override
    public int indexOf(Object o) {
        return elements.indexOf(o);
    }

    @Override
    public int lastIndexOf(Object o) {
        return elements.lastIndexOf(o);
    }

    @Override
    public @NotNull ListIterator<@Nullable JsonValue> listIterator() {
        return elements.listIterator();
    }

    @Override
    public @NotNull ListIterator<@Nullable JsonValue> listIterator(int index) {
        return elements.listIterator(index);
    }

    @Override
    public @NotNull List<@Nullable JsonValue> subList(int fromIndex, int toIndex) {
        return elements.subList(fromIndex, toIndex);
    }
    //</editor-fold>

    /**
     * {@return the element at the specified position in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified position is neither {@code null} nor an instance of {@link JsonString}.
     */
    public @Nullable JsonString getString(int index) {
        return get(index, JsonString.class);
    }

    /**
     * {@return the element at the specified position in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified position is neither {@code null} nor an instance of {@link JsonNumber}.
     */
    public @Nullable JsonNumber getNumber(int index) {
        return get(index, JsonNumber.class);
    }

    /**
     * {@return the element at the specified position in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified position is neither {@code null} nor an instance of {@link JsonBoolean}.
     */
    public @Nullable JsonBoolean getBoolean(int index) {
        return get(index, JsonBoolean.class);
    }

    /**
     * {@return the element at the specified position in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified position is neither {@code null} nor an instance of {@link JsonArray}.
     */
    public @Nullable JsonArray getArray(int index) {
        return get(index, JsonArray.class);
    }

    /**
     * {@return the element at the specified position in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified position is neither {@code null} nor an instance of {@link JsonObject}.
     */
    public @Nullable JsonObject getObject(int index) {
        return get(index, JsonObject.class);
    }

    private <T extends JsonValue> @Nullable T get(int index, @NotNull Class<T> type) {
        return type.cast(elements().get(index));
    }

    @Override
    public @NotNull String toString() {
        return elements.toString();
    }

    @Override
    public @NotNull String toJsonString() {
        var joiner = new StringJoiner(", ", "[", "]");
        for (JsonValue element : elements) {
            joiner.add(JsonValue.toJsonString(element));
        }
        return joiner.toString();
    }

    @Override
    public @NotNull String toPrettyJsonString() {
        if (size() < 2) return toJsonString();

        var out = new StringJoiner(",\n", "[\n", "\n]");
        elements().forEach(e -> out.add(indent(JsonValue.toPrettyJsonString(e))));
        return out.toString();
    }

    private static @NotNull String indent(@NotNull String string) {
        if (string.isEmpty()) return "";

        StringBuilder out = new StringBuilder();

        Iterator<String> it = string.lines().iterator();
        while (it.hasNext()) {
            out.append("    ").append(it.next());
            if (it.hasNext()) out.append('\n');
        }

        return out.toString();
    }
}
