package eu.jonahbauer.json;

import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Function;

/**
 * Java representation of a JSON object.
 * @param entries the object's entries
 */
@SuppressWarnings("unused")
public record JsonObject(@NotNull Map<@NotNull String, @Nullable JsonValue> entries) implements JsonValue, Map<@NotNull String, @Nullable JsonValue> {

    public JsonObject {
        Objects.requireNonNull(entries, "entries");
        if (!Util.isImmutable(entries)) {
            if (entries instanceof SequencedMap<?, ?>) {
                var map = new LinkedHashMap<String, JsonValue>();
                entries.forEach((key, value) -> map.put(Objects.requireNonNull(key, "key"), value));
                entries = Collections.unmodifiableSequencedMap(map);
            } else {
                var map = new HashMap<String, JsonValue>();
                entries.forEach((key, value) -> map.put(Objects.requireNonNull(key, "key"), value));
                entries = Collections.unmodifiableMap(map);
            }
        }
    }

    /**
     * Creates a {@link JsonObject} from the given map, using {@link JsonValue#valueOf(Object)} to convert the values to
     * {@link JsonValue}s. The keys are expected to be instances of {@link CharSequence}.
     * @param map a map
     * @return a {@link JsonObject}
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonObject valueOf(@Nullable Map<?, ?> map) {
        if (map == null) return null;
        if (map instanceof JsonObject json) return json;
        
        var out = new LinkedHashMap<String, JsonValue>();
        map.forEach((key, value) -> out.put(keyOf(key), JsonValue.valueOf(value)));
        return new JsonObject(out);
    }

    private static @NotNull String keyOf(@Nullable Object object) {
        return switch (object) {
            case CharSequence chars -> chars.toString();
            case null -> throw new JsonProcessingException("cannot convert null to json key");
            default -> throw new JsonProcessingException(STR."cannot convert object of type \{object.getClass()} to json key");
        };
    }

    //<editor-fold desc="Map" defaultstate="collapsed">
    @Override
    public int size() {
        return entries.size();
    }

    @Override
    public boolean isEmpty() {
        return entries.isEmpty();
    }

    @Override
    public boolean containsKey(Object key) {
        Objects.requireNonNull(key, "key");
        return entries.containsKey(key);
    }

    @Override
    public boolean containsValue(Object value) {
        return entries.containsValue(value);
    }

    @Override
    public @Nullable JsonValue get(Object key) {
        Objects.requireNonNull(key, "key");
        return entries.get(key);
    }

    @Override
    public @Nullable JsonValue put(@NotNull String key, @Nullable JsonValue value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public @Nullable JsonValue remove(Object key) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void putAll(@NotNull Map<? extends @NotNull String, ? extends @Nullable JsonValue> m) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }

    @Override
    public @NotNull Set<@NotNull String> keySet() {
        return entries.keySet();
    }

    @Override
    public @NotNull Collection<@Nullable JsonValue> values() {
        return entries.values();
    }

    @Override
    public @NotNull Set<Entry<@NotNull String, @Nullable JsonValue>> entrySet() {
        return entries.entrySet();
    }
    //</editor-fold>

    //<editor-fold desc="typed get(...)" defaultstate="collapsed">
    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is neither {@code null} nor an instance of {@link JsonString}
     */
    public @Nullable JsonString getString(@NotNull String key) {
        return get(key, JsonString.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is neither {@code null} nor an instance of {@link JsonNumber}
     */
    public @Nullable JsonNumber getNumber(@NotNull String key) {
        return get(key, JsonNumber.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is neither {@code null} nor an instance of {@link JsonBoolean}
     */
    public @Nullable JsonBoolean getBoolean(@NotNull String key) {
        return get(key, JsonBoolean.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is neither {@code null} nor an instance of {@link JsonArray}
     */
    public @Nullable JsonArray getArray(@NotNull String key) {
        return get(key, JsonArray.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is neither {@code null} nor an instance of {@link JsonObject}
     */
    public @Nullable JsonObject getObject(@NotNull String key) {
        return get(key, JsonObject.class);
    }

    private <T extends JsonValue> @Nullable T get(@NotNull String key, @NotNull Class<T> type) {
        return type.cast(entries().get(key));
    }
    //</editor-fold>

    //<editor-fold desc="typed getOrDefault(...)" defaultstate="collapsed">
    // getStringOrDefault(String, JsonString) omitted because JsonString implements CharSequence

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonString#valueOf(CharSequence) JsonString.valueOf}{@code (defaultValue)}
     * if this object contains no mapping for the key or key is mapped to an object that is neither {@code null} nor an
     * instance of {@link JsonString}.
     * @return the value to which the specified key is mapped, or
     *         {@link JsonString#valueOf(CharSequence) JsonString.valueOf}{@code (defaultValue)}
     *         if this object contains no mapping for the key or key is mapped to an object that is neither {@code null}
     *         nor an instance of {@link JsonString}
     */
    public @Nullable JsonString getStringOrDefault(@NotNull String key, @Nullable CharSequence defaultValue) {
        return getOrDefault(key, defaultValue, JsonString.class, JsonString::valueOf);
    }

    /**
     * Returns the value to which the specified key is mapped, or {@code defaultValue} if this object contains no mapping
     * for the key or key is mapped to an object that is neither {@code null} nor an instance of {@link JsonNumber}.
     * @return the value to which the specified key is mapped, or {@code defaultValue} if this object contains no mapping
     *         for the key or key is mapped to an object that is neither {@code null} nor an instance of
     *         {@link JsonNumber}
     */
    public @Nullable JsonNumber getNumberOrDefault(@NotNull String key, @Nullable JsonNumber defaultValue) {
        return getOrDefault(key, defaultValue, JsonNumber.class, Function.identity());
    }

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonNumber#valueOf(Number) JsonNumber.valueOf}{@code (defaultValue)}
     * if this object contains no mapping for the key or key is mapped to an object that is neither {@code null} nor an
     * instance of {@link JsonNumber}.
     * @return the value to which the specified key is mapped, or
     *         {@link JsonNumber#valueOf(Number) JsonNumber.valueOf}{@code (defaultValue)}
     *         if this object contains no mapping for the key or key is mapped to an object that is neither {@code null}
     *         nor an instance of {@link JsonNumber}
     */
    public @Nullable JsonNumber getNumberOrDefault(@NotNull String key, @Nullable Number defaultValue) {
        return getOrDefault(key, defaultValue, JsonNumber.class, JsonNumber::valueOf);
    }

    /**
     * Returns the value to which the specified key is mapped, or {@code defaultValue} if this object contains no mapping
     * for the key or key is mapped to an object that is neither {@code null} nor an instance of {@link JsonBoolean}.
     * @return the value to which the specified key is mapped, or {@code defaultValue} if this object contains no mapping
     *         for the key or key is mapped to an object that is neither {@code null} nor an instance of
     *         {@link JsonBoolean}
     */
    public @Nullable JsonBoolean getBooleanOrDefault(@NotNull String key, @Nullable JsonBoolean defaultValue) {
        return getOrDefault(key, defaultValue, JsonBoolean.class, Function.identity());
    }

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonBoolean#valueOf(Boolean) JsonBoolean.valueOf}{@code (defaultValue)}
     * if this object contains no mapping for the key or key is mapped to an object that is neither {@code null} nor an
     * instance of {@link JsonBoolean}.
     * @return the value to which the specified key is mapped, or
     *         {@link JsonBoolean#valueOf(Boolean) JsonBoolean.valueOf}{@code (defaultValue)}
     *         if this object contains no mapping for the key or key is mapped to an object that is neither {@code null}
     *         nor an instance of {@link JsonBoolean}
     */
    public @Nullable JsonBoolean getBooleanOrDefault(@NotNull String key, @Nullable Boolean defaultValue) {
        return getOrDefault(key, defaultValue, JsonBoolean.class, JsonBoolean::valueOf);
    }
    
    // getArrayOrDefault(String, JsonArray) omitted because JsonArray implements List<?>

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonArray#valueOf(List) JsonArray.valueOf}{@code (defaultValue)}
     * if this object contains no mapping for the key or key is mapped to an object that is neither {@code null} nor an
     * instance of {@link JsonArray}.
     * @return the value to which the specified key is mapped, or
     *         {@link JsonArray#valueOf(List) JsonArray.valueOf}{@code (defaultValue)}
     *         if this object contains no mapping for the key or key is mapped to an object that is neither {@code null}
     *         nor an instance of {@link JsonArray}
     */
    public @Nullable JsonArray getArrayOrDefault(@NotNull String key, @Nullable List<?> defaultValue) {
        return getOrDefault(key, defaultValue, JsonArray.class, JsonArray::valueOf);
    }

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonArray#valueOf(Object...) JsonArray.valueOf}{@code (defaultValue)}
     * if this object contains no mapping for the key or key is mapped to an object that is neither {@code null} nor an
     * instance of {@link JsonArray}.
     * @return the value to which the specified key is mapped, or
     *         {@link JsonArray#valueOf(Object...) JsonArray.valueOf}{@code (defaultValue)}
     *         if this object contains no mapping for the key or key is mapped to an object that is neither {@code null}
     *         nor an instance of {@link JsonArray}
     */
    public @Nullable JsonArray getArrayOrDefault(@NotNull String key, @Nullable Object @Nullable... defaultValue) {
        return getOrDefault(key, defaultValue, JsonArray.class, JsonArray::valueOf);
    }

    // getObjectOrDefault(String, JsonObject) omitted because JsonObject implements Map<?, ?>

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonObject#valueOf(Map) JsonObject.valueOf}{@code (defaultValue)}
     * if this object contains no mapping for the key or key is mapped to an object that is neither {@code null} nor an
     * instance of {@link JsonObject}.
     * @return the value to which the specified key is mapped, or
     *         {@link JsonObject#valueOf(Map) JsonObject.valueOf}{@code (defaultValue)}
     *         if this object contains no mapping for the key or key is mapped to an object that is neither {@code null}
     *         nor an instance of {@link JsonObject}
     */
    public @Nullable JsonObject getObjectOrDefault(@NotNull String key, @Nullable Map<?, ?> defaultValue) {
        return getOrDefault(key, defaultValue, JsonObject.class, JsonObject::valueOf);
    }
    
    private <T extends JsonValue, S> @Nullable T getOrDefault(@NotNull String key, @Nullable S defaultValue, @NotNull Class<T> type, @NotNull Function<S, T> converter) {
        JsonValue value = entries.get(key);
        if (type.isInstance(value)) {
            return type.cast(value);
        } else if (value == null && containsKey(key)) {
            return null;
        } else {
            return converter.apply(defaultValue);
        }
    }
    //</editor-fold>

    @Override
    public @NotNull String toString() {
        StringJoiner out = new StringJoiner(",", "{", "}");
        entries.forEach((key, value) -> out.add(JsonString.quote(key) + ":" + value));
        return out.toString();
    }

    @Override
    public @NotNull String toPrettyJsonString() {
        StringJoiner out = new StringJoiner(",\n    ", "{\n    ", "\n}");
        entries.forEach((key, value) -> out.add(JsonString.quote(key) + ": " + indent(JsonValue.toPrettyJsonString(value))));
        return out.toString();
    }

    private static @NotNull String indent(@NotNull String string) {
        if (string.isEmpty()) return "";

        StringBuilder out = new StringBuilder();

        Iterator<String> it = string.lines().iterator();
        out.append(it.next());

        while (it.hasNext()) {
            String next = it.next();
            out.append('\n').append(it.hasNext() ? "        " : "    ").append(next);
        }

        return out.toString();
    }
}
