package eu.jonahbauer.json;

import eu.jonahbauer.json.parser.JsonConversionException;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.Serializable;
import java.util.List;
import java.util.Map;

public sealed interface JsonValue extends Serializable permits JsonObject, JsonArray, JsonBoolean, JsonNumber, JsonString {

    /**
     * {@return the JSON representation of this value}
     */
    default @NotNull String toJsonString() {
        return toString();
    }

    /**
     * {@return a prettified JSON representation of this value}
     */
    default @NotNull String toPrettyJsonString() {
        return toJsonString();
    }

    default @Nullable JsonValue select(@NotNull JsonPath path) {
        return path.select(this);
    }

    default @Nullable JsonValue select(@NotNull String path) {
        return JsonPath.parse(path).select(this);
    }

    /**
     * {@return the JSON representation of the given value}
     * @param value a {@link JsonValue} (possibly {@code null})
     */
    static @NotNull String toJsonString(@Nullable JsonValue value) {
        return value == null ? "null" : value.toJsonString();
    }

    /**
     * {@return the prettified JSON representation of the given value}
     * @param value a {@link JsonValue} (possibly {@code null})
     */
    static @NotNull String toPrettyJsonString(@Nullable JsonValue value) {
        return value == null ? "null" : value.toPrettyJsonString();
    }

    /**
     * Tries to convert {@code object} to a {@link JsonValue}:
     * <ul>
     *     <li>{@link JsonValue}s and {@code null} are simple returned</li>
     *     <li>{@link Number}s are converted with {@link JsonNumber#valueOf(Number)}</li>
     *     <li>{@link CharSequence}s are converted with {@link JsonString#valueOf(CharSequence)}</li>
     *     <li>{@link Boolean}s are converted with {@link JsonBoolean#valueOf(Boolean)}</li>
     *     <li>{@link List}s are converted with {@link JsonArray#valueOf(List)}</li>
     *     <li>
     *         {@code Object[]}, {@code int[]} and {@code double[]} are converted with
     *         {@link JsonArray#valueOf(Object...)}, {@link JsonArray#valueOf(int...)} and
     *         {@link JsonArray#valueOf(double...)} resp.
     *     </li>
     *     <li>{@link Map}s are converted with {@link JsonObject#valueOf(Map)}</li>
     * </ul>
     * @param object an object
     * @return a {@link JsonValue} representing the object
     * @throws JsonConversionException if the object cannot be converted to a {@link JsonValue}
     */
    @Contract("null -> null; !null -> !null")
    static @Nullable JsonValue valueOf(@Nullable Object object) {
        return switch (object) {
            case JsonValue json -> json;
            case Number number -> JsonNumber.valueOf(number);
            case CharSequence chars -> JsonString.valueOf(chars);
            case Boolean bool -> JsonBoolean.valueOf(bool);
            case List<?> list -> JsonArray.valueOf(list);
            case Object[] array -> JsonArray.valueOf(array);
            case int[] array -> JsonArray.valueOf(array);
            case double[] array -> JsonArray.valueOf(array);
            case Map<?, ?> map -> JsonObject.valueOf(map);
            case null -> null;
            default -> throw new JsonConversionException(STR."cannot convert object of type \{object.getClass()} to json value");
        };
    }
}
