package eu.jonahbauer.json.parser;

import eu.jonahbauer.json.JsonObject;
import eu.jonahbauer.json.JsonProcessingException;
import eu.jonahbauer.json.JsonValue;
import org.example.json.*;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.*;

class JsonProcessorTest {

    @Test
    @SuppressWarnings("StringOperationCanBeSimplified")
    void processWithoutPlaceholders() {
        JsonValue value = JsonProcessor.JSON."""
        {
            "string": "string",
            "number": 1337,
            "true": true,
            "false": false,
            "null": null,
            "array": [1, "2", true, false, null]
        }
        """;

        assertInstanceOf(JsonObject.class, value);
        JsonObject object = (JsonObject) value;

        assertEquals(JsonValue.valueOf("string"), object.get("string"));
        assertEquals(JsonValue.valueOf(1337), object.get("number"));
        assertEquals(JsonValue.valueOf(true), object.get("true"));
        assertEquals(JsonValue.valueOf(false), object.get("false"));
        assertEquals(JsonValue.valueOf(null), object.get("null"));
        assertEquals(JsonValue.valueOf(Arrays.asList(1, "2", true, false, null)), object.get("array"));
    }

    @Test
    @SuppressWarnings("TrailingWhitespacesInTextBlock")
    void processWithValuePlaceholder() {
        JsonValue value = JsonProcessor.JSON."""
        {
            "string": \{"string"},
            "number": \{1337},
            "true": \{true},
            "false": \{false},
            "null": \{null},
            "array": \{Arrays.asList(1, "2", true, false, null)}
        }
        """;

        assertInstanceOf(JsonObject.class, value);
        JsonObject object = (JsonObject) value;

        assertEquals(JsonValue.valueOf("string"), object.get("string"));
        assertEquals(JsonValue.valueOf(1337), object.get("number"));
        assertEquals(JsonValue.valueOf(true), object.get("true"));
        assertEquals(JsonValue.valueOf(false), object.get("false"));
        assertEquals(JsonValue.valueOf(null), object.get("null"));
        assertEquals(JsonValue.valueOf(Arrays.asList(1, "2", true, false, null)), object.get("array"));
    }

    @Test
    void processWithKeyPlaceholder() {
        String key = "foo";
        JsonValue value = JsonProcessor.JSON."""
        {
            \{key}: "value"
        }
        """;

        assertInstanceOf(JsonObject.class, value);
        JsonObject object = (JsonObject) value;

        assertEquals(JsonValue.valueOf("value"), object.get(key));
    }

    @Test
    void processWithPlaceholderInInvalidPosition() {
        assertThrows(JsonProcessingException.class, () -> {
            var _ = JsonProcessor.JSON."{\"key\" \{"value"}}";
        }).printStackTrace(System.out);
    }

    @Test
    void processWithPlaceholderInStringLiteral() {
        Object object = new Object();
        JsonValue value = JsonProcessor.JSON."{\"key\": \"Hello \{object}!\"}";
        assertEquals(JsonValue.valueOf(Map.of(
                "key", "Hello " + object + "!"
        )), value);
    }

    @Test
    void processWithPlaceholderInKeyStringLiteral() {
        Object object = new Object();
        JsonValue value = JsonProcessor.JSON."{\"key-\{object}\": \"value\"}";
        assertEquals(JsonValue.valueOf(Map.of(
                "key-" + object, "value"
        )), value);
    }


    @ParameterizedTest(name = "{0}")
    @MethodSource("parameters")
    void suite(String name) throws IOException {
        var path = "/nst/JsonTestSuite/" + name;
        Boolean expected = switch (name.charAt(0)) {
            case 'i' -> null;
            case 'y' -> true;
            case 'n' -> false;
            default -> throw new IllegalArgumentException();
        };

        try (var in = Objects.requireNonNull(JsonProcessorTest.class.getResourceAsStream(path))) {
            String input = new String(in.readAllBytes(), StandardCharsets.UTF_8);

            if (expected == Boolean.FALSE) {
                assertThrows(JsonProcessingException.class, () -> JsonProcessor.JSON.process(StringTemplate.of(input))).printStackTrace(System.out);
            } else if (expected == Boolean.TRUE) {
                Assertions.assertDoesNotThrow(() -> JsonProcessor.JSON.process(StringTemplate.of(input)));
            } else {
                try {
                    JsonProcessor.JSON.process(StringTemplate.of(input));
                    System.out.println("accepted");
                } catch (JsonProcessingException ex) {
                    System.out.println("rejected");
                    ex.printStackTrace(System.out);
                }
            }
        }
    }

    static Stream<Arguments> parameters() throws IOException {
        List<Arguments> filenames = new ArrayList<>();

        try (
                InputStream in = Objects.requireNonNull(JsonProcessorTest.class.getResource("/nst/JsonTestSuite")).openStream();
                BufferedReader br = new BufferedReader(new InputStreamReader(in))
        ) {
            String resource;
            while ((resource = br.readLine()) != null) {
                filenames.add(Arguments.of(resource));
            }
        }

        return filenames.stream();
    }
}
